import os
from pathlib import Path
import xmlrpc.client
import pandas as pd
from datetime import datetime, timedelta

def api_params_func(test_db: bool = False) -> dict:

    api_url = os.environ.get('ODOO_URL_API')
    api_db = os.environ.get('ODOO_DB_API')
    api_test_db = os.environ.get('ODOO_DB_PRUEBA_API')
    api_username = os.environ.get('ODOO_USERNAME_API')
    api_clave = os.environ.get('ODOO_CLAVE_API')


    api_params = {}
    if test_db:
        api_params['api_db'] = api_test_db
    else:
        api_params['api_db'] = api_db


    common = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/common')
    uid = common.authenticate(api_params['api_db'], api_username, api_clave, {})
    models = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/object')


    api_params['api_clave'] = api_clave
    api_params['api_uid'] = uid
    api_params['api_models'] = models

    return api_params

def search_pay_func(mes: int) -> list[str]:
    
    if type(mes) != int or mes < 1 or mes > 12:
        raise Exception (f'El mes es incorrecto. El párametro "mes" debe ser un número entero entre 1 y 12. Escribiste: {mes}')
    

    param_dia_ini = datetime(2024, mes, 1)
    param_dia_fin = datetime(2024, mes + 1, 1) - timedelta(days= 1)

    search_pay_acc = [
        "&", "&", "&",
            ("partner_type", "=", "customer"),
            ("is_internal_transfer", "=", False),
            ("state", "=", "posted"),
        "&",
            ("date", ">=", param_dia_ini.strftime('%Y-%m-%d')),
            ("date", "<=", param_dia_fin.strftime('%Y-%m-%d')),
        ]


    param_dia_ini_pos = param_dia_ini + timedelta(hours=7)
    param_dia_fin_pos = param_dia_fin + timedelta(hours=31) - timedelta(seconds= 1)
    
    search_pay_pos = [
        "&",
            ("payment_date", ">=", param_dia_ini_pos.strftime('%Y-%m-%d %H:%M:%S')),
            ("payment_date", "<=", param_dia_fin_pos.strftime('%Y-%m-%d %H:%M:%S')),
        ]
    

    return search_pay_acc, search_pay_pos

def api_call_pay_acc_func(api_params: dict, search_pay: list[str] ) -> list[dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']


    search_pay_acc = search_pay[0]


    pay_acc_fields = [
                    'name',
                    'date',
                    'partner_id',
                    'amount',
                    'ref',
                    'reconciled_invoice_ids',
                    'pos_session_id'
                    ]

    pay_acc_ids = models.execute_kw(api_db, uid, api_clave, 'account.payment', 'search', [search_pay_acc])
    pay_acc_json = models.execute_kw(api_db, uid, api_clave, 'account.payment', 'read', [pay_acc_ids], {'fields': pay_acc_fields})

    
    return pay_acc_json

def pay_acc_df_func(pay_acc_json: list[dict]) -> list[pd.DataFrame, int]:
    
    data_pay_acc = []
    data_fact_ids = []

    for pay in pay_acc_json:
        if not pay['pos_session_id']:
            new = {}
            new['id'] = pay['id']
            new['name'] = pay['name']
            new['date'] = pay['date']
            new['partner_id'] = pay['partner_id']
            new['amount'] = pay['amount']
            new['ref'] = pay['ref'] if pay['ref'] else pd.NA
            new['pay_fact_docs'] = pay['reconciled_invoice_ids'] if pay['reconciled_invoice_ids'] else pd.NA
            
            data_fact_ids += pay['reconciled_invoice_ids']

            data_pay_acc.append(new)

    pay_acc_df = pd.DataFrame(data_pay_acc)


    pay_acc_df['date'] = pd.to_datetime(pay_acc_df['date'], format='%Y-%m-%d')
    
    return pay_acc_df, data_fact_ids

def api_call_pay_pos_func(api_params: dict, search_pay: list[str] ) -> list[dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']


    search_pay_pos = search_pay[1]


    pay_pos_fields = [
                    'pos_order_id',
                    'payment_date',
                    'amount',
                    'payment_method_id'
                    ]

    pay_pos_ids = models.execute_kw(api_db, uid, api_clave, 'pos.payment', 'search', [search_pay_pos])
    pay_pos_json = models.execute_kw(api_db, uid, api_clave, 'pos.payment', 'read', [pay_pos_ids], {'fields': pay_pos_fields})
    

    return pay_pos_json

def pay_pos_df_func(pay_pos_json: list[dict]) -> list[pd.DataFrame, int]:
    
    data_pay_pos = []
    data_pos_ids = set()

    for pay in pay_pos_json:
        new = {}
        new['pay_pos_id'] = pay['id']
        new['pos_order_id'] = pay['pos_order_id'][0]
        new['pos_order_name'] = pay['pos_order_id'][1]
        new['date'] = pay['payment_date']
        new['amount'] = pay['amount']
        new['payment_method'] = pay['payment_method_id'][1]
    
        data_pos_ids.add(pay['pos_order_id'][0])
        data_pay_pos.append(new)

    pay_pos_df = pd.DataFrame(data_pay_pos)

    pay_pos_df['date'] = pd.to_datetime(pay_pos_df['date'], format='%Y-%m-%d %H:%M:%S')
    
    return pay_pos_df, list(data_pos_ids)

def api_call_pos_order_func(api_params: dict, data_pos_ids: list[int] ) -> list[dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']

    pos_order_fields = [
                    'name',
                    'account_move',
                    'partner_id',
                    'session_id',
                    'amount_total'
                    ]

    pos_order_json = models.execute_kw(api_db, uid, api_clave, 'pos.order', 'read', [data_pos_ids], {'fields': pos_order_fields})

    return pos_order_json

def pos_order_func(pos_order_json: list[dict], data_fact_ids: list[int]) -> list[list[int]]:
    
    data_pos_amount_cero = [] 
  
    for pos in pos_order_json:
        
        if pos['account_move']:
            data_fact_ids.append(pos['account_move'][0])

        if pos['amount_total'] == 0:
            data_pos_amount_cero.append(pos)


    pos_amount_cero_df = pd.DataFrame(data_pos_amount_cero)
    
    if not pos_amount_cero_df.empty:
        pos_amount_cero_df['session_id'] = pos_amount_cero_df['session_id'].str.get(1)
        
    
    return data_fact_ids, pos_amount_cero_df

def api_call_acccount_docs_func(api_params: dict, data_fact_ids: list[int] ) -> list[dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']

    data_fact_fields = [
                    'name',
                    'partner_id',
                    'date',
                    'invoice_payments_widget',
                    'amount_total',
                    'amount_residual'
                    ]

    fact_doc_json = models.execute_kw(api_db, uid, api_clave, 'account.move', 'read', [list(set(data_fact_ids))], {'fields': data_fact_fields})

    return fact_doc_json

def pay_fact_df_func(mes: int, fact_doc_json: dict) -> pd.DataFrame:
    
    data_pay_fact = []

    for fact in fact_doc_json:
        if fact['invoice_payments_widget']:
            for pay in fact['invoice_payments_widget']['content']:
                if 'Facturas' not in pay['journal_name'] and datetime.strptime(pay['date'], '%Y-%m-%d').month == mes:
                    new = {}
                    new['fact_doc_id'] = fact['id']
                    new['fact_doc_name'] = fact['name']
                    new['fact_doc_cliente'] = fact['partner_id'][0]
                    new['fact_doc_date'] = fact['date']
                    new['fact_doc_total'] = fact['amount_total']
                    new['fact_doc_deuda'] = fact['amount_residual']
                    new['pay_journal'] = pay['journal_name']

                    if 'PdV ' in pay['ref']:
                        i = pay['ref'].find('PdV ')
                        new['ref'] = pay['ref'][i:i+12]
                    elif 'Shop' in pay['ref']:
                        i = pay['ref'].find('Shop')
                        new['ref'] = pay['ref'][i:i+9]
                    else:
                        new['ref'] = pd.NA

                    new['pay_amount'] = pay['amount']
                    new['pay_date'] = pay['date']
                    new['pay_pos'] = pay['pos_payment_name'] if pay['pos_payment_name'] else pd.NA
        
                    data_pay_fact.append(new)

    pay_fact_df = pd.DataFrame(data_pay_fact)

    pay_fact_df.loc[~pay_fact_df['pay_journal'].str.contains('Punto De Venta'), ['ref']] = pd.NA

    pay_fact_df['fact_doc_date'] = pd.to_datetime(pay_fact_df['fact_doc_date'], format='%Y-%m-%d')
    pay_fact_df['pay_date'] = pd.to_datetime(pay_fact_df['pay_date'], format='%Y-%m-%d')

    return pay_fact_df

def pagos_faltantes_check_func(pay_acc_df: pd.DataFrame, pos_amount_cero_df: pd.DataFrame, pay_pos_df: pd.DataFrame) -> bool:
    
    pagos_faltantes = []

    if not pos_amount_cero_df.empty:
        acc_cobranza_pdv_df = pay_acc_df[(~pay_acc_df['ref'].isna()) & (pay_acc_df['ref'].str[:1] != 'F')]
        pdv_cobranza_df = pos_amount_cero_df.merge(pay_pos_df, how='left', left_on='id', right_on='pos_order_id')

        for i in range(len(pdv_cobranza_df)):
            mini_df = acc_cobranza_pdv_df.loc[
                (acc_cobranza_pdv_df['ref'].str.contains(pdv_cobranza_df['name'].iloc[i])) 
                | (acc_cobranza_pdv_df['ref'].str.contains(pdv_cobranza_df['session_id'].iloc[i])) 
                & (acc_cobranza_pdv_df['amount'] == pdv_cobranza_df['amount'].iloc[i])
                & (acc_cobranza_pdv_df['partner_id'].str.get(0) == pdv_cobranza_df['partner_id'].iloc[i][0])]

            if mini_df.empty:
                new = {}
                new['pos_doc_id'] = pdv_cobranza_df.iloc[i]['id']
                new['pos_doc_name'] = pdv_cobranza_df.iloc[i]['name']
                new['cliente'] = pdv_cobranza_df.iloc[i]['partner_id'][1]
                new['monto'] = pdv_cobranza_df.iloc[i]['amount']
                new['fecha'] = pdv_cobranza_df.iloc[i]['date']- timedelta(hours=7)

                pagos_faltantes.append(new)

    pagos_faltantes_df = pd.DataFrame(pagos_faltantes)

    if pagos_faltantes_df.empty:
        return True

    archivo = 'Cobranza_PdV_Faltante'
    path = Path.home().joinpath(f'Desktop/{archivo}.xlsx')
    pagos_faltantes_df.to_excel(path)

def cobranza_mes_func(mes: int, test_db: bool = False) -> pd.DataFrame:

    api_params = api_params_func(test_db)
    search_pay = search_pay_func(mes)

    pay_acc_json = api_call_pay_acc_func(api_params, search_pay)
    pay_acc_df, data_fact_ids2 = pay_acc_df_func(pay_acc_json)

    pay_pos_json = api_call_pay_pos_func(api_params, search_pay)
    pay_pos_df, data_pos_ids = pay_pos_df_func(pay_pos_json)

    pos_order_json = api_call_pos_order_func(api_params, data_pos_ids)
    data_fact_ids, pos_amount_cero_df = pos_order_func(pos_order_json, data_fact_ids2)

    pagos_faltantes_check = pagos_faltantes_check_func(pay_acc_df, pos_amount_cero_df, pay_pos_df)

    if pagos_faltantes_check:
        fact_doc_json = api_call_acccount_docs_func(api_params, data_fact_ids)
        pay_fact_df = pay_fact_df_func(mes, fact_doc_json)
        
        return pay_fact_df

    else:
        print('Corrige los pagos de cobranza PdV faltantes, el archivo está en tu escritorio')



import os
from pathlib import Path
import xmlrpc.client
import numpy as np
import pandas as pd
from sqlalchemy import create_engine, Engine


def api_params_func(test_db: bool = False) -> dict:

    api_url = os.environ.get('ODOO_URL_API')
    api_db = os.environ.get('ODOO_DB_API')
    api_test_db = os.environ.get('ODOO_DB_PRUEBA_API')
    api_username = os.environ.get('ODOO_USERNAME_API')
    api_clave = os.environ.get('ODOO_CLAVE_API')


    api_params = {}
    if test_db:
        api_params['api_db'] = api_test_db
    else:
        api_params['api_db'] = api_db


    common = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/common')
    uid = common.authenticate(api_params['api_db'], api_username, api_clave, {})
    models = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/object')


    api_params['api_clave'] = api_clave
    api_params['api_uid'] = uid
    api_params['api_models'] = models

    return api_params

def search_costo_ventas_func(mes: int) -> list[str]:
    
    if type(mes) != int or mes < 1 or mes > 12:
        raise Exception (f'El mes es incorrecto. El párametro "mes" debe ser un número entero entre 1 y 12. Escribiste: {mes}')
    
    param_dia_hr_ini = pd.Timestamp(2024, mes, 1) + pd.Timedelta(hours=7)
    
    if mes == 12:
        param_dia_hr_fin = pd.Timestamp(2025, 1, 1) + pd.Timedelta(hours=7) - pd.Timedelta(seconds= 1)
    else:
        param_dia_hr_fin = pd.Timestamp(2024, mes+1, 1) + pd.Timedelta(hours=7) - pd.Timedelta(seconds= 1)

    search_costo_ventas = [
        "&", "&",
            ("state", "in", ["purchase", "done"]),
            ("date_approve", ">=", param_dia_hr_ini.strftime('%Y-%m-%d %H:%M:%S')),
            ("date_approve", "<=", param_dia_hr_fin.strftime('%Y-%m-%d %H:%M:%S')),
        ]

    return search_costo_ventas

def _get_df_from_excel(file_name:str, file_location:str) -> pd.DataFrame:

    file_path_str = str(Path().cwd().parent.parent.joinpath(f'data/{file_location}/{file_name}'))
    df = pd.read_excel(file_path_str, dtype_backend='numpy_nullable')

    return df

def prov_locales_df_from_excel() -> list[pd.DataFrame]:
    
    file_name = 'proveedores_oficiales.xlsx'
    file_location = 'compras'

    df = _get_df_from_excel(file_name, file_location)

    prov_oficiales = df.loc[df['oficial'] == 1][['partner_id', 'partner_name']]
    prov_locales = df.loc[df['oficial'] == 0][['partner_id', 'partner_name']]

    return prov_oficiales, prov_locales

def clientes_comisiones_df_from_excel(mes:int) -> list[pd.DataFrame]:
    
    file_name = 'clientes_comisiones.xlsx'
    file_location = 'comisiones'

    df = _get_df_from_excel(file_name, file_location)

    if not df[df['partner_id_x'].duplicated()].empty:
        print(f'Hay "partner_id_x" (cientes comisionistas) que están duplicados en el archivo {file_name}. Esto en el merge va a generar una doble línea. ¡Corrígelo!')
    
    
    dia_ini = pd.Timestamp(2024, mes, 1)
    
    if mes == 12:
        dia_fin = pd.Timestamp(2024, 12, 31)
    else:
        dia_fin = pd.Timestamp(2024, mes+1, 1) - pd.Timedelta(days=1)


    clientes_comisiones = df.loc[
                                (df['fecha_alta'] <= dia_ini) 
                                & (
                                    (df['fecha_baja'] >= dia_fin) 
                                    | (df['fecha_baja'].isna()) 
                                )
                            ]
    
    return clientes_comisiones

def _ultimo_costo_sae_df_from_excel() -> list[pd.DataFrame]:
    
    file_name = 'ultimo_costo_sae.xlsx'
    file_location = 'costo_ventas'

    df = _get_df_from_excel(file_name, file_location)

    return df

def _get_db_engine(db_mode:str) -> Engine:
    
    if db_mode.lower() == 'local':

        db_file_path_str = str(Path().cwd().parent.parent.joinpath(f'data/comisiones.db'))
        engine = create_engine(f'sqlite:///{db_file_path_str}')

        return engine

    else:
        raise Exception (f'Sólo existe la base de datos "Local"')

def get_dfs_from_database(db_mode: str, mes:int) -> list[pd.DataFrame]:

    engine = _get_db_engine(db_mode)
    mes_anterior = mes -1


    if mes != 1:
    
        with engine.connect() as conn, conn.begin():
            
            try:
                ventas = pd.read_sql_table(f'ventas_{mes}_2024', conn, dtype_backend='numpy_nullable')
            except:
                print(f'No se encontró la tabla ventas_{mes}_2024 en la base de datos {db_mode}')

            try:
                ultimo_costo = pd.read_sql_table(f'ultimo_costo_{mes_anterior}_2024', conn, dtype_backend='numpy_nullable')
            except:
                print(f'No se encontró la tabla ultimo_costo_{mes_anterior}_2024 en la base de datos {db_mode}')

            try:
                compras_especiales_sin_usar = pd.read_sql_table(f'compras_especiales_sin_usar_{mes_anterior}_2024', conn, dtype_backend='numpy_nullable')
            except:
                print(f'No se encontró la tabla compras_especiales_sin_usar_{mes_anterior}_2024 en la base de datos {db_mode}')

        engine.dispose()


        if type(ventas) != pd.DataFrame or type(ultimo_costo) != pd.DataFrame or type(compras_especiales_sin_usar) != pd.DataFrame:
            return None, None, None


    else:

        with engine.connect() as conn, conn.begin():
            
            try:
                ventas = pd.read_sql_table(f'ventas_{mes}_2024', conn, dtype_backend='numpy_nullable')
            except:
                print(f'¡Cuidado, no continúes! No se encontró la tabla ventas_{mes}_2024 en la base de datos {db_mode}')
                ventas = None

            ultimo_costo = _ultimo_costo_sae_df_from_excel()
            compras_especiales_sin_usar = None

        engine.dispose()


    return ventas, ultimo_costo, compras_especiales_sin_usar

def correcciones_df_from_excel(mes:int) ->pd.DataFrame:
    
    file_name = f'correcciones_costo_ventas_{mes}_2024.xlsx'
    file_location = 'costo_ventas'

    try:
        correcciones = _get_df_from_excel(file_name, file_location)
    
    except:
        print(f'No existe el archivo {file_name}')
        return None, []
    
    line_ids_correcciones = list(
                correcciones.loc[
                        (~correcciones['line_id_correct'].isna())
                        & (correcciones['line_id_correct'] != 'SAE')
                    ]
                    ['line_id_correct']
                    .astype(int)
                    .sort_values()
                )

    return correcciones, line_ids_correcciones

def api_call_purchase_doc_func(api_params: dict, search_costo_ventas: list[str], line_ids_correcciones:list ) -> list[int, dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']

    purchase_doc_fields = [
                    'name',
                    'state',
                    'partner_id',
                    'partner_ref',
                    'date_approve',
                    'x_fecha_factura',
                    'user_id',
                    'create_uid'
                    ]

    purchase_doc_ids1 = models.execute_kw(api_db, uid, api_clave, 'purchase.order', 'search', [search_costo_ventas])
    
    if len(line_ids_correcciones) != 0:
        purchase_doc_ids2 = models.execute_kw(api_db, uid, api_clave, 'purchase.order', 'search', [[("order_line", "in", line_ids_correcciones)]])
    
    else:
        purchase_doc_ids2 = set()
    
    purchase_doc_ids = list(set(purchase_doc_ids2) | set(purchase_doc_ids1))
    ids_doc_correcciones_dif_mes = list(set(purchase_doc_ids2) - set(purchase_doc_ids1))

    purchase_doc_json = models.execute_kw(api_db, uid, api_clave, 'purchase.order', 'read', [purchase_doc_ids], {'fields': purchase_doc_fields})
     
    return purchase_doc_ids, purchase_doc_json, ids_doc_correcciones_dif_mes

def purchase_doc_func(purchase_doc_json: list[dict], prov_oficiales:pd.DataFrame) -> pd.DataFrame:

    purchase_doc_data = []

    for compra in purchase_doc_json:
        new = {}
        new['order_id'] = compra['id']
        new['order_name'] = compra['name']
        new['order_state'] = compra['state']
        new['order_date'] = compra['date_approve'] if compra['date_approve'] else pd.NA
        new['partner_id'] = compra['partner_id'][0]
        new['partner_name'] = compra['partner_id'][1]
        new['partner_fact_ref'] = compra['partner_ref']
        new['partner_fact_date'] = compra['x_fecha_factura'] if compra['x_fecha_factura'] else pd.NA
        new['capturista'] = compra['create_uid'][1] if compra['create_uid'] else pd.NA
        new['vendedora'] = compra['user_id'][1] if compra['user_id'] else pd.NA

        purchase_doc_data.append(new)

    compras_doc = pd.DataFrame(purchase_doc_data)
    compras_doc['order_date'] = pd.to_datetime(compras_doc['order_date'], format='%Y-%m-%d %H:%M:%S')
    compras_doc['partner_fact_date'] = pd.to_datetime(compras_doc['partner_fact_date'], format='%Y-%m-%d')
    compras_doc['oficial'] = compras_doc['partner_id'].isin(prov_oficiales['partner_id'])

    return compras_doc


def api_call_purchase_line_func(api_params: dict, purchase_doc_ids: list[int]) -> list[dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']

    purchase_line_fields = [
                        'order_id',
                        'date_approve',
                        'partner_id',
                        'product_id',
                        'product_qty',
                        'price_unit_discounted'
                        ]

    purchase_line_ids = models.execute_kw(api_db, uid, api_clave, 'purchase.order.line', 'search', [[("order_id.id", "in", purchase_doc_ids)]])
    purchase_line_json = models.execute_kw(api_db, uid, api_clave, 'purchase.order.line', 'read', [purchase_line_ids], {'fields': purchase_line_fields})
    
    return purchase_line_json

def purchase_line_func(purchase_line_json: list[dict]) -> pd.DataFrame:
    
    purchase_line_data = []

    for line in purchase_line_json:
        new = {}
        new['line_id'] = line['id']
        new['order_id'] = line['order_id'][0]
        new['order_name'] = line['order_id'][1]
        new['order_date'] = line['date_approve'] if line['date_approve'] else pd.NA
        new['partner_id'] = line['partner_id'][0]
        new['partner_name'] = line['partner_id'][1]
        new['product_id_pp'] = line['product_id'][0]
        new['product_name'] = line['product_id'][1]
        new['product_qty'] = line['product_qty']
        new['product_cost'] = line['price_unit_discounted']
        
        purchase_line_data.append(new)

    compras_line = pd.DataFrame(purchase_line_data)
    compras_line['order_date'] = pd.to_datetime(compras_line['order_date'], format='%Y-%m-%d %H:%M:%S')

    return compras_line

def compras_odoo_func(compras_doc:pd.DataFrame, compras_line:pd.DataFrame, line_ids_correcciones:list, ids_doc_correcciones_dif_mes:list) -> pd.DataFrame:
    
    compras_odoo_total = pd.merge(
                    compras_line,
                    compras_doc[['order_id', 'partner_fact_ref', 'partner_fact_date', 'capturista', 'vendedora']], 
                    how='left', 
                    on='order_id'
                )

    compras_odoo_total['order_date'] = compras_odoo_total['order_date'].dt.normalize()

    cols_to_Int64 = ['line_id', 'order_id', 'partner_id', 'product_id_pp']
    compras_odoo_total[cols_to_Int64] = compras_odoo_total[cols_to_Int64].astype('Int64')

    compras_odoo_total['product_qty'] = compras_odoo_total['product_qty'].astype('Float64')
    compras_odoo_total['vendedora'] = compras_odoo_total['vendedora'].convert_dtypes()

    compras_odoo = compras_odoo_total[~compras_odoo_total['order_id'].isin(ids_doc_correcciones_dif_mes)]
    compras_line_id_corrected = compras_odoo_total[
                                        compras_odoo_total['line_id'].isin(line_ids_correcciones)
                                    ].rename(columns={'partner_id': 'partner_id_y',
                                                      'partner_name': 'partner_name_y',
                                                      'product_name': 'product_name_y'})

    return compras_odoo, compras_line_id_corrected

def lista_capturistas_func(mes:int) -> list:
    
    lista_capturistas = [
            'Elsa Ivette Diaz Leyva',
            'Alexa Yadira Mazariegos Zunun',
            'Dulce Guadalupe Pedroza Valenzuela',
            'Mariana Araceli Carvajal Flores',
            'Rosario Martinez Zarate'
        ]

    if mes <= 6:
        return lista_capturistas
        
    else:
        return lista_capturistas + ['Patricia Flores Pantaleón']

def division_compras_especiales_func(compras_odoo:pd.DataFrame, compras_especiales_sin_usar:pd.DataFrame, lista_capturistas:list, prov_oficiales:pd.DataFrame) -> list[pd.DataFrame]:

    # Concatenación de compras especiales del mes anterior con las compras de este mes de odoo
    total_compras_odoo = pd.concat([
                compras_odoo,
                compras_especiales_sin_usar
        ])

    # Compras especiales del mes en curso que además no son de proveedores oficiales
    compras_especiales = total_compras_odoo.loc[
                (~total_compras_odoo['vendedora'].isin(lista_capturistas))
                & (~total_compras_odoo['partner_id'].isin(prov_oficiales['partner_id']))
        ]

    # Resto de compras del mes en curso, es decir, compras de venta normal
    compras_no_especiales = total_compras_odoo.loc[
                ~total_compras_odoo['line_id'].isin(compras_especiales['line_id'])
        ]

    return total_compras_odoo, compras_especiales, compras_no_especiales

def match_1_merge_func(ventas:pd.DataFrame, compras_especiales:pd.DataFrame) -> list[pd.DataFrame]:

    match_1_repetidos = pd.DataFrame([], columns=['line_id','fact_line_id'])

    compras_especiales_while = compras_especiales.copy()
    ventas_while = ventas.copy()
    repetidos_while = True

    while repetidos_while:
        
        match_1_merge = (
            pd.merge_asof(
                left = compras_especiales_while.sort_values('order_date'), 
                right = ventas_while.sort_values('invoice_date'),
                
                left_by = ['product_id_pp', 'vendedora', 'product_qty'], 
                right_by = ['product_id', 'salesperson_name', 'quantity'], 
                
                left_on = 'order_date', 
                right_on = 'invoice_date', 

                direction = 'nearest',
                tolerance = pd.Timedelta(days=3))
        )

        ids_ventas_repetidas_por_corregir = match_1_merge.loc[
                                                    (~match_1_merge['fact_line_id'].isna()) 
                                                    & (match_1_merge['fact_line_id'].duplicated())
                                            ]['fact_line_id'].drop_duplicates()
        
        if not ids_ventas_repetidas_por_corregir.empty:
            repetidos_mini_df = match_1_merge.loc[match_1_merge['fact_line_id'].isin(ids_ventas_repetidas_por_corregir)].sort_values('line_id')
            repetidos_mini_df['diff'] = abs(repetidos_mini_df['invoice_date'] - repetidos_mini_df['order_date'])

            for id in ids_ventas_repetidas_por_corregir:
                match_line_to_keep = repetidos_mini_df.loc[repetidos_mini_df['fact_line_id'] == id].sort_values('diff').reset_index().loc[:0, ['line_id', 'fact_line_id']]

                if match_1_repetidos.empty:
                    match_1_repetidos = match_line_to_keep.copy()
                else:
                    match_1_repetidos = pd.concat([match_1_repetidos, match_line_to_keep])

            compras_especiales_while = compras_especiales_while[~compras_especiales_while['line_id'].isin(match_1_repetidos['line_id'])]
            ventas_while = ventas_while[~ventas_while['fact_line_id'].isin(match_1_repetidos['fact_line_id'])]


        if ids_ventas_repetidas_por_corregir.empty:        
            print('Terminó el ciclo while')
            repetidos_while = False


    return match_1_merge, match_1_repetidos

def match_1_results_func(match_1_merge:pd.DataFrame, match_1_repetidos:pd.DataFrame, ventas:pd.DataFrame, compras_especiales:pd.DataFrame) -> list[pd.DataFrame]:

    # Match resultantes del 1er match.
    match_1 = match_1_merge.loc[~match_1_merge['fact_line_id'].isna(), ['line_id', 'fact_line_id']]

    # Después de correr el 1er match, las ventas restantes
    ventas_after_match_1 = ventas.loc[
                                    (~ventas['fact_line_id'].isin(match_1['fact_line_id']))
                                    & (~ventas['fact_line_id'].isin(match_1_repetidos['fact_line_id']))
                                ]

    # Después de correr el 1er match, las compras que tienen un product_id que sí existe en las ventas restantes.
    compras_especiales_after_match_1 = compras_especiales.loc[
                                                (~compras_especiales['line_id'].isin(match_1['line_id']))
                                                & (~compras_especiales['line_id'].isin(match_1_repetidos['line_id']))
                                                & (compras_especiales['product_id_pp'].isin(ventas_after_match_1['product_id']))
                                            ]

    # Después de correr el 1er match, resto de las compras especiales. Estan no tienen un product_id que existe en las ventas restantes y no se pueden merchar.
    compras_especiales_after_match_1_sin_linea_venta = compras_especiales.loc[
                                                                (~compras_especiales['line_id'].isin(match_1['line_id']))
                                                                & (~compras_especiales['line_id'].isin(match_1_repetidos['line_id']))
                                                                & (~compras_especiales['product_id_pp'].isin(ventas_after_match_1['product_id']))
                                                            ]
    
    return match_1, ventas_after_match_1, compras_especiales_after_match_1, compras_especiales_after_match_1_sin_linea_venta

def match_2_merge_func(ventas_after_match_1:pd.DataFrame, compras_especiales_after_match_1:pd.DataFrame) -> list[pd.DataFrame]:

    # Varias ventas para una sóla compra.
    match_2 = pd.DataFrame([], columns=['line_id','fact_line_id'])

    for i in range(len(compras_especiales_after_match_1)):
        
        linea_compra = compras_especiales_after_match_1.sort_values('order_date').iloc[i]

        mini_df = ventas_after_match_1.loc[
                    (~ventas_after_match_1['fact_line_id'].isin(match_2['fact_line_id']))
                    & (ventas_after_match_1['salesperson_name'] == linea_compra['vendedora'])
                    & (ventas_after_match_1['product_id'] == linea_compra['product_id_pp'])
                ]
        
        if not mini_df.empty:
            df_copia = mini_df.copy()
            df_copia['diff'] = abs(df_copia['invoice_date'] - linea_compra['order_date'])
            df_sort = df_copia.sort_values('diff').reset_index()
            df_sort['cumsum'] = df_sort['quantity'].cumsum()
            index = df_sort[df_sort['cumsum'] == linea_compra['product_qty']].index[0] if not df_sort[df_sort['cumsum'] == linea_compra['product_qty']].index.empty else None
            if index != None:
                df_sort.loc[:index , 'line_id'] = linea_compra['line_id']
                df_to_keep = df_sort.loc[~df_sort['line_id'].isna()]

                if match_2.empty:
                    match_2 = df_to_keep[['line_id','fact_line_id']].copy()
                else:
                    match_2 = pd.concat([
                                match_2, 
                                df_to_keep[['line_id','fact_line_id']],
                            ])


    return match_2

def match_2_results_func(match_2:pd.DataFrame, ventas_after_match_1:pd.DataFrame, compras_especiales_after_match_1:pd.DataFrame) -> list[pd.DataFrame]:

    # Después de correr el 2do match, las ventas restantes
    ventas_after_match_2 = ventas_after_match_1.loc[
                                    ~ventas_after_match_1['fact_line_id'].isin(match_2['fact_line_id'])
                                ]

    # Después de correr el 2do match, las compras que restan y que sí tendrán venta a la cual mercharse.
    compras_especiales_after_match_2 = compras_especiales_after_match_1.loc[
                                                (~compras_especiales_after_match_1['line_id'].isin(match_2['line_id']))
                                                & (compras_especiales_after_match_1['product_id_pp'].isin(ventas_after_match_2['product_id']))
                                            ]

    # Después de correr el 2do match, resto de las compras especiales. Estan no tienen un product_id que existe en las ventas restantes y no se pueden merchar.
    compras_especiales_after_match_2_sin_linea_venta = compras_especiales_after_match_1.loc[
                                                                (~compras_especiales_after_match_1['line_id'].isin(match_2['line_id']))
                                                                & (~compras_especiales_after_match_1['product_id_pp'].isin(ventas_after_match_2['product_id']))
                                                            ]
    
    return ventas_after_match_2, compras_especiales_after_match_2, compras_especiales_after_match_2_sin_linea_venta

def match_3_merge_func(ventas_after_match_2:pd.DataFrame, compras_especiales_after_match_2:pd.DataFrame) -> list[pd.DataFrame]:

    # Varias compras para una sóla venta.
    match_3 = pd.DataFrame([], columns=['line_id','fact_line_id'])
    match_3_to_errase = pd.DataFrame([], columns=['line_id','fact_line_id'])

    ventas_inside_compras_especiales_after_match_2 = ventas_after_match_2[
                                                            ventas_after_match_2['product_id'].isin(compras_especiales_after_match_2['product_id_pp'])
                                                        ]

    for i in range(len(ventas_inside_compras_especiales_after_match_2)):
        
        linea_venta = ventas_inside_compras_especiales_after_match_2.sort_values('invoice_date').iloc[i]

        mini_df = compras_especiales_after_match_2.loc[
                    (~compras_especiales_after_match_2['line_id'].isin(match_3['line_id']))
                    & (compras_especiales_after_match_2['vendedora'] == linea_venta['salesperson_name'])
                    & (compras_especiales_after_match_2['product_id_pp'] == linea_venta['product_id'])
                ]
            

        if not mini_df.empty:
        
            df_copia = mini_df.copy()
            df_copia['diff'] = abs(df_copia['order_date'] - linea_venta['invoice_date'])
            df_sort = df_copia.sort_values('diff').reset_index()
            df_sort['cumsum'] = df_sort['product_qty'].cumsum()
            index = df_sort[df_sort['cumsum'] == linea_venta['quantity']].index[0] if not df_sort[df_sort['cumsum'] == linea_venta['quantity']].index.empty else None
            
            if index != None:
                df_sort.loc[:index , 'fact_line_id'] = linea_venta['fact_line_id']
                
                #Esta línea es para poder detectar la línea de compra con el costo mayor y dejarla para la línea de venta
                df_to_keep = (
                        df_sort.loc[
                                ~df_sort['fact_line_id'].isna()
                            ]
                            .sort_values('product_cost', ascending=False)
                            .reset_index()
                        )

                if match_3.empty:
                    match_3 = df_to_keep.loc[:0, ['line_id', 'fact_line_id']].copy()
                    match_3_to_errase = df_to_keep[['line_id', 'fact_line_id']].copy()

                else:
                    match_3 = pd.concat([
                                match_3, 
                                df_to_keep.loc[:0, ['line_id', 'fact_line_id']],
                            ])
                    match_3_to_errase = pd.concat([
                                match_3_to_errase, 
                                df_to_keep[['line_id', 'fact_line_id']],
                            ])


    return match_3, match_3_to_errase

def match_3_results_func(match_3_to_errase:pd.DataFrame, ventas_after_match_2:pd.DataFrame, compras_especiales_after_match_2:pd.DataFrame) -> list[pd.DataFrame]:

    # Después de correr el 3er match, las ventas restantes
    ventas_after_match_3 = ventas_after_match_2.loc[
                                    ~ventas_after_match_2['fact_line_id'].isin(match_3_to_errase['fact_line_id'])
                                ]

    # Después de correr el 3er match, las compras que restan y que sí tendrán venta a la cual mercharse.
    compras_especiales_after_match_3 = compras_especiales_after_match_2.loc[
                                                (~compras_especiales_after_match_2['line_id'].isin(match_3_to_errase['line_id']))
                                                & (compras_especiales_after_match_2['product_id_pp'].isin(ventas_after_match_3['product_id']))
                                            ]

    # Después de correr el 3er match, resto de las compras especiales. Estan no tienen un product_id que existe en las ventas restantes y no se pueden merchar.
    compras_especiales_after_match_3_sin_linea_venta = compras_especiales_after_match_2.loc[
                                                                (~compras_especiales_after_match_2['line_id'].isin(match_3_to_errase['line_id']))
                                                                & (~compras_especiales_after_match_2['product_id_pp'].isin(ventas_after_match_3['product_id']))
                                                            ]
    
    return ventas_after_match_3, compras_especiales_after_match_3, compras_especiales_after_match_3_sin_linea_venta

def match_4_merge_func(ventas_after_match_3:pd.DataFrame, compras_especiales_after_match_3:pd.DataFrame) -> list[pd.DataFrame]:

    match_4_merge = (
        pd.merge_asof(
            left = ventas_after_match_3.sort_values('invoice_date'), 
            right = compras_especiales_after_match_3.sort_values('order_date'),
            
            left_by = ['product_id'], 
            right_by = ['product_id_pp'], 
            
            left_on = 'invoice_date', 
            right_on = 'order_date', 

            direction = 'nearest',
        )
    )

    return match_4_merge

def match_4_results_func(match_4_merge:pd.DataFrame, ventas_after_match_3:pd.DataFrame, compras_especiales_after_match_3:pd.DataFrame) -> list[pd.DataFrame]:

    # Match resultantes del 4to match.
    match_4 = match_4_merge.loc[~match_4_merge['line_id'].isna(), ['line_id', 'fact_line_id']]

    # Después de correr el 1er match, las ventas restantes
    ventas_after_match_4 = ventas_after_match_3.loc[
                                    ~ventas_after_match_3['fact_line_id'].isin(match_4['fact_line_id'])
                                ]

    # Después de correr el 4to match, las compras que tienen un product_id que sí existe en las ventas restantes.
    compras_especiales_after_match_4 = compras_especiales_after_match_3.loc[
                                                (~compras_especiales_after_match_3['line_id'].isin(match_4['line_id']))
                                                & (compras_especiales_after_match_3['product_id_pp'].isin(ventas_after_match_4['product_id']))
                                            ]

    # Después de correr el 4to match, resto de las compras especiales. Estan no tienen un product_id que existe en las ventas restantes y no se pueden merchar.
    compras_especiales_after_match_4_sin_linea_venta = compras_especiales_after_match_3.loc[
                                                                (~compras_especiales_after_match_3['line_id'].isin(match_4['line_id']))
                                                                & (~compras_especiales_after_match_3['product_id_pp'].isin(ventas_after_match_4['product_id']))
                                                            ]
    
    return match_4, ventas_after_match_4, compras_especiales_after_match_4, compras_especiales_after_match_4_sin_linea_venta

def match_compras_especiales_func(match_1:pd.DataFrame, match_1_repetidos:pd.DataFrame, match_2:pd.DataFrame, match_3:pd.DataFrame, match_4:pd.DataFrame) -> pd.DataFrame:

    match_compras_especiales = pd.concat(
                                    [
                                        match_1, 
                                        match_1_repetidos if not match_1_repetidos.empty else None, 
                                        match_2, 
                                        match_3, 
                                        match_4
                                    ]
                                )

    return match_compras_especiales

def costo_ventas_a_func(match_compras_especiales:pd.DataFrame, ventas:pd.DataFrame, total_compras_odoo:pd.DataFrame) -> pd.DataFrame:

    costo_ventas_a = (
        match_compras_especiales.merge(
            ventas,
            how='left',
            on='fact_line_id'
        ).merge(
            total_compras_odoo,
            how='left',
            on='line_id'
        )
    )

    return costo_ventas_a

def costo_ventas_b_func(ventas_after_match_4:pd.DataFrame, compras_no_especiales:pd.DataFrame, ultimo_costo:pd.DataFrame) -> pd.DataFrame:

    costo_ventas_b = (
        pd.merge_asof(
            left = ventas_after_match_4.sort_values('invoice_date'),
            right = pd.concat([
                            compras_no_especiales, 
                            ultimo_costo,
                    ]).sort_values('order_date'), 
            
            left_by = 'product_id', 
            right_by = 'product_id_pp', 
            
            left_on = 'invoice_date', 
            right_on = 'order_date', 

            direction = 'backward')
    )

    return costo_ventas_b

def costo_ventas_func(costo_ventas_a:pd.DataFrame, costo_ventas_b:pd.DataFrame) -> pd.DataFrame:

    costo_ventas = pd.concat([costo_ventas_a, costo_ventas_b])

    costo_ventas['fact_line_id'] = costo_ventas['fact_line_id'].astype('Int64')
    costo_ventas['line_id'] = costo_ventas['line_id'].astype('Int64')

    return costo_ventas

def costo_ventas_corregido_func(correcciones:pd.DataFrame|None, compras_line_id_corrected:pd.DataFrame | None, costo_ventas:pd.DataFrame) -> pd.DataFrame:

    costo_ventas_corregido = costo_ventas.copy()
    costo_ventas_corregido[['is_verified', 'is_corrected','comentario_correct']] = pd.NA
    
    if type(correcciones) != pd.DataFrame:
        print('No se aplicadó ninguna corrección al costo de venta')
    
    else:
        # fact_line_ids verificados
        ids_verified = correcciones[~correcciones['is_verified'].isna()]
        
        for i in range(len(ids_verified)):
            costo_ventas_corregido.loc[
                        costo_ventas_corregido['fact_line_id'] == ids_verified.iloc[i]['fact_line_id']
                        , 
                        ['is_verified', 'comentario_correct']
                    ] = True, ids_verified.iloc[i]['comentario_correct']

        
        # fact_line_ids que van a ser corregidos
        ids_corrected = correcciones[correcciones['is_verified'].isna()]
        
        costo_ventas_corregido.loc[
                costo_ventas_corregido['fact_line_id'].isin(ids_corrected['fact_line_id'])
                , 
                ['is_corrected']
            ] = True
        
        
        # line_id desde el SAE
        ids_sae = correcciones[correcciones['line_id_correct'] == 'SAE']
        
        cols_to_correct_sae = ['line_id', 'order_id', 'partner_id_y', 'partner_name_y', 'product_qty',
        'partner_fact_ref', 'partner_fact_date', 'capturista', 'vendedora']

        for i in range(len(ids_sae)):
            costo_ventas_corregido.loc[
                        costo_ventas_corregido['fact_line_id'] == ids_sae.iloc[i]['fact_line_id']
                        , 
                        cols_to_correct_sae
                    ] = pd.NA
            
            costo_ventas_corregido.loc[
                        costo_ventas_corregido['fact_line_id'] == ids_sae.iloc[i]['fact_line_id']
                        , 
                        ['order_name', 'order_date', 'product_cost', 'comentario_correct']
                    ] = 'SAE', pd.Timestamp(2023, 12, 31), ids_sae.iloc[i]['costo_correct'], ids_sae.iloc[i]['comentario_correct']
        

        # line_id incorrectas, cambiar a nueva line_id correcta
        ids_new_line = correcciones[
                                (correcciones['line_id_correct'] != 'SAE')
                                & (~correcciones['line_id_correct'].isna())
                            ]
        
        # Check para ver que todas la line_id_corrected escritas en el archivo de correcciones se encontraron en Odoo
        check_line_id_correct = ids_new_line['line_id_correct'].astype(int)
        check = check_line_id_correct[~check_line_id_correct.isin(compras_line_id_corrected['line_id'])]

        if not check.empty:
            print(f'\nLos siguientes "line_id_correct" en el archivo de correcciones no se encontraron en Odoo:')
            print(f'    {check.values}')
            return

        cols_to_correct_new_line_id = ['line_id', 'order_id', 'order_name', 'order_date',
        'partner_id_y', 'partner_name_y', 'product_id_pp', 'product_name_y', 'product_qty',
        'partner_fact_ref', 'partner_fact_date', 'capturista', 'vendedora']

        for i in range(len(ids_new_line)):
            costo_ventas_corregido.loc[
                        costo_ventas_corregido['fact_line_id'] == ids_new_line.iloc[i]['fact_line_id']
                        , 
                        cols_to_correct_new_line_id
                    ] = compras_line_id_corrected.loc[
                                compras_line_id_corrected['line_id'] == int(ids_new_line.iloc[i]['line_id_correct'])
                                , cols_to_correct_new_line_id
                            ].values
            
            costo_ventas_corregido.loc[
                    costo_ventas_corregido['fact_line_id'] == ids_new_line.iloc[i]['fact_line_id']
                    , 
                    ['product_cost', 'comentario_correct']
                ] = ids_new_line.iloc[i]['costo_correct'], ids_new_line.iloc[i]['comentario_correct']
            
            
    # Se procede a generar las columnas de utilidades y margenes con las correcciones generadas
    costo_ventas_corregido['cost_subtotal'] = costo_ventas_corregido['product_cost'] * costo_ventas_corregido['quantity']
    costo_ventas_corregido['utilidad_partida_$'] = costo_ventas_corregido['price_subtotal'] - costo_ventas_corregido['cost_subtotal']
    costo_ventas_corregido['utilidad_%'] = ((costo_ventas_corregido['price_subtotal'] / (costo_ventas_corregido['cost_subtotal']) - 1) * 100).round(2)
    costo_ventas_corregido['margen_contribución_%'] = (costo_ventas_corregido['utilidad_partida_$'] / costo_ventas_corregido['price_subtotal'] * 100).round(2)

    # Corregir los margenes de contribución infinitos, ya que se dividieron entre precio cero
    costo_ventas_corregido.replace([np.inf, -np.inf], pd.NA, inplace=True)

    return costo_ventas_corregido

def costo_ventas_after_comisiones_clientes_func(costo_ventas_corregido:pd.DataFrame, clientes_comisiones:pd.DataFrame) -> pd.DataFrame:

    if type(costo_ventas_corregido) != pd.DataFrame:
        return


    costo_ventas_after_comisiones_clientes = costo_ventas_corregido.merge(
                                                        clientes_comisiones[['partner_id_x', '%_comision_cliente']],
                                                        how = 'left',
                                                        on = 'partner_id_x'
                                                    )

    costo_ventas_after_comisiones_clientes['cost_cliente_comision'] = costo_ventas_after_comisiones_clientes['price_subtotal'] * costo_ventas_after_comisiones_clientes['%_comision_cliente'] / 100
    
    costo_ventas_after_comisiones_clientes['cost_subtotal_after_cc'] = costo_ventas_after_comisiones_clientes['cost_subtotal'] + costo_ventas_after_comisiones_clientes['cost_cliente_comision']
    costo_ventas_after_comisiones_clientes['utilidad_partida_$_after_cc'] = costo_ventas_after_comisiones_clientes['price_subtotal'] - costo_ventas_after_comisiones_clientes['cost_subtotal_after_cc']
    costo_ventas_after_comisiones_clientes['utilidad_%_after_cc'] = ((costo_ventas_after_comisiones_clientes['price_subtotal'] / (costo_ventas_after_comisiones_clientes['cost_subtotal_after_cc']) - 1) * 100).round(2)
    costo_ventas_after_comisiones_clientes['margen_contribución_%_after_cc'] = (costo_ventas_after_comisiones_clientes['utilidad_partida_$_after_cc'] / costo_ventas_after_comisiones_clientes['price_subtotal'] * 100).round(2)


    # Corregir utilidad y márgen pd.NA resultante de los clientes que no llevan comisión
    index_clientes_no_comision = costo_ventas_after_comisiones_clientes.loc[
                costo_ventas_after_comisiones_clientes['%_comision_cliente'].isna(),
            ].index

    costo_ventas_after_comisiones_clientes.loc[
            index_clientes_no_comision,
            [
                'cost_subtotal_after_cc',
                'utilidad_partida_$_after_cc',
                'utilidad_%_after_cc',
                'margen_contribución_%_after_cc'
            ]
        ] = costo_ventas_corregido[[
                'cost_subtotal',
                'utilidad_partida_$',
                'utilidad_%',
                'margen_contribución_%'
            ]].iloc[index_clientes_no_comision].values

    return costo_ventas_after_comisiones_clientes

def nuevas_compras_especiales_sin_usar_func(compras_especiales_after_match_1_sin_linea_venta:pd.DataFrame, compras_especiales_after_match_2_sin_linea_venta:pd.DataFrame, compras_especiales_after_match_3_sin_linea_venta:pd.DataFrame, compras_especiales_after_match_4_sin_linea_venta:pd.DataFrame, compras_odoo:pd.DataFrame,) -> pd.DataFrame:

    ids_compras_especiales_sin_usar = pd.concat(
            [
                compras_especiales_after_match_1_sin_linea_venta['line_id'],
                compras_especiales_after_match_2_sin_linea_venta['line_id'],
                compras_especiales_after_match_3_sin_linea_venta['line_id'],
                compras_especiales_after_match_4_sin_linea_venta['line_id'],
            ]
        )

    nuevas_compras_especiales_sin_usar = compras_odoo[
                                    compras_odoo['line_id'].isin(ids_compras_especiales_sin_usar)
                                ]

    return nuevas_compras_especiales_sin_usar

def nuevo_ultimo_costo_func(ultimo_costo:pd.DataFrame, compras_especiales:pd.DataFrame, compras_no_especiales:pd.DataFrame) -> pd.DataFrame:

    nuevo_ultimo_costo = (
            pd.concat(
                    [
                    ultimo_costo, 
                    compras_no_especiales,
                    compras_especiales[
                                ~compras_especiales['product_id_pp'].isin(compras_no_especiales['product_id_pp'])
                            ]
                    ]
                )
            .sort_values('order_date', ascending=False)
            .groupby('product_id_pp')
            .first()
            .reset_index()
        )

    return nuevo_ultimo_costo

def checks_costo_ventas_func(
            mes:int,
            compras_doc:pd.DataFrame,
            prov_locales:pd.DataFrame,
            prov_oficiales:pd.DataFrame,
            compras_especiales_after_match_4:pd.DataFrame,
            match_compras_especiales:pd.DataFrame,
            match_3_to_errase:pd.DataFrame,
            match_3:pd.DataFrame,
            compras_especiales_after_match_1_sin_linea_venta:pd.DataFrame,
            compras_especiales_after_match_2_sin_linea_venta:pd.DataFrame,
            compras_especiales_after_match_3_sin_linea_venta:pd.DataFrame,
            compras_especiales_after_match_4_sin_linea_venta:pd.DataFrame,
            compras_especiales:pd.DataFrame,
            costo_ventas_corregido:pd.DataFrame,
            ventas:pd.DataFrame
    ) -> bool:
    
    # Línea para comprobrar que el 100% de los proveedores de Odoo están calificados en la lista de proveedores oficiales
    check1 = (compras_doc[~compras_doc['partner_id'].isin(pd.concat([prov_locales, prov_oficiales])['partner_id'])]).drop_duplicates('partner_id')


    # Check donde se verifíca que ya no hay más compras especiales por merchar, es decir, ya no se ocupa un match_5.
    check2 = compras_especiales_after_match_4.empty


    # Check donde todas las líneas compras especiales están unidas a líneas de ventas
    all_line_ids_processed = (
            pd.concat(
                [
                    match_compras_especiales['line_id'],
                    match_3_to_errase.loc[~match_3_to_errase['line_id'].isin(match_3), 'line_id'],
                    compras_especiales_after_match_1_sin_linea_venta['line_id'],
                    compras_especiales_after_match_2_sin_linea_venta['line_id'],
                    compras_especiales_after_match_3_sin_linea_venta['line_id'],
                    compras_especiales_after_match_4_sin_linea_venta['line_id'],
                ]
            )
            .drop_duplicates()
            .sort_values()
            .astype('Int64')
            .reset_index()
            ['line_id']
        )

    all_line_ids_to_process = (
            compras_especiales['line_id']
            .sort_values()
            .reset_index()
            ['line_id']
        )

    check3 = all_line_ids_processed.equals(all_line_ids_to_process)


    # Check para ver el tamaño del costo de ventas generado vs las ventas del mes
    check4 = len(costo_ventas_corregido) == len(ventas)


    # Check para ver que no existen lineas de venta fuera del costo de ventas.
    check5 = ventas[~ventas['fact_line_id'].isin(costo_ventas_corregido['fact_line_id'])].empty


    # Check para que todas las líneas de venta del costo de ventas tienen un costo.
    productos_aceptables_sin_costo_ventas = ['Anticipo', 
                                            'Anticipo (PdV)', 
                                            'Servicios de Facturación'
                                            ]

    check6 = costo_ventas_corregido.loc[
            (costo_ventas_corregido['order_name'].isna())
            & (~costo_ventas_corregido['product_name_x'].isin(productos_aceptables_sin_costo_ventas))
        ]


    # Check para ver todas las líneas que tienen una utilidad negativa o en ceros.
    check7 = costo_ventas_corregido.loc[
            (costo_ventas_corregido['move_type'] != 'out_refund')
            & (costo_ventas_corregido['quantity'] != 0)
            & (costo_ventas_corregido['utilidad_partida_$'] <= 0)
            & (costo_ventas_corregido['is_verified'].isna())
            & (costo_ventas_corregido['is_corrected'].isna())
        ].sort_values('utilidad_partida_$')


    # Check para ver todas las líneas que tienen una utilidad mayor a 200%.
    check8 = costo_ventas_corregido.loc[
            (costo_ventas_corregido['move_type'] != 'out_refund')
            & (costo_ventas_corregido['utilidad_%'] >= 200)
            & (costo_ventas_corregido['is_verified'].isna())
            & (costo_ventas_corregido['is_corrected'].isna())
        ].sort_values('utilidad_%', ascending=False)


    # Checar todos los cheks
    all_checks = (
                check1.empty
                and check2
                and check3
                and check4
                and check5
                and check6.empty
                and check7.empty
                and check8.empty
    )

    if all_checks:
        print('¡Todos los checks salieron correctos!')
        return all_checks

    else:
        print(f'Fallaron los cheks:')

        not check1.empty and print('   -Check1: Hay proveedores no calificados en la lista de proveedores oficiales.')
        not check2 and print('   -Check2: Hay compras especiales remanentes después del último match.')
        not check3 and print('   -Check3: Hay líneas de compra especiales que aun no se han procesado.')
        not check4 and print('   -Check4: El tamaño del costo de ventas es diferente al tamaño de las ventas iniciales.')
        not check5 and print('   -Check5: El costo de ventas no contempla unas líneas de venta del dataframe de ventas iniciales.')
        not check6.empty and print(f'   -Check6: Hay {len(check6)} productos que no tienen costo de ventas.')
        not check7.empty and print(f'   -Check7: Hay {len(check7)} líneas de venta con utilidades negativas.')
        not check8.empty and print(f'   -Check8: Hay {len(check8)} líneas de venta con utilidades mayores al 200%')


        # Escribe el archivo .xlsx en el escritorio para la corrección de los checks 7 y 8.
        if not check6.empty or not check7.empty or not check8.empty:
            archivo = f'lineas_por_corregir_costo_ventas_{mes}_2024'
            path = Path.home().joinpath(f'Desktop/{archivo}.xlsx')
            writer = pd.ExcelWriter(path, engine="openpyxl")

            cols_to_write_corrections = ['line_id_correct', 'costo_correct']
            cols_to_excel = list(costo_ventas_corregido.columns) + cols_to_write_corrections

            if not check6.empty:
                check6.loc[:, cols_to_write_corrections] = pd.NA
                check6[cols_to_excel].to_excel(writer, sheet_name='prod_sin_costo')
            
            if not check7.empty:
                check7.loc[:, cols_to_write_corrections] = pd.NA
                check7[cols_to_excel].to_excel(writer, sheet_name='negativos')

            if not check8.empty:
                check8.loc[:, cols_to_write_corrections] = pd.NA
                check8[cols_to_excel].to_excel(writer, sheet_name='exagerados')


            writer.close()
            writer.handles = None

            print(f'\nSe generó el archivo {archivo}.xlsx y se guardó en el escritorio.')


def costo_ventas_mes_func(mes: int, db_mode:str) -> pd.DataFrame:

    api_params = api_params_func()
    search_costo_ventas = search_costo_ventas_func(mes)

    prov_oficiales, prov_locales = prov_locales_df_from_excel()
    clientes_comisiones = clientes_comisiones_df_from_excel(mes)
    ventas, ultimo_costo, compras_especiales_sin_usar = get_dfs_from_database(db_mode, mes)

    correcciones, line_ids_correcciones = correcciones_df_from_excel(mes)

    purchase_doc_ids, purchase_doc_json, ids_doc_correcciones_dif_mes = api_call_purchase_doc_func(api_params, search_costo_ventas, line_ids_correcciones)
    purchase_line_json = api_call_purchase_line_func(api_params, purchase_doc_ids)

    compras_doc = purchase_doc_func(purchase_doc_json, prov_oficiales)
    compras_line = purchase_line_func(purchase_line_json)
    compras_odoo, compras_line_id_corrected = compras_odoo_func(compras_doc, compras_line, line_ids_correcciones, ids_doc_correcciones_dif_mes)

    lista_capturistas = lista_capturistas_func(mes)
    total_compras_odoo, compras_especiales, compras_no_especiales = division_compras_especiales_func(compras_odoo, compras_especiales_sin_usar, lista_capturistas, prov_oficiales)

    match_1_merge, match_1_repetidos = match_1_merge_func(ventas, compras_especiales)
    match_1, ventas_after_match_1, compras_especiales_after_match_1, compras_especiales_after_match_1_sin_linea_venta = match_1_results_func(match_1_merge, match_1_repetidos, ventas, compras_especiales)

    match_2 = match_2_merge_func(ventas_after_match_1, compras_especiales_after_match_1)
    ventas_after_match_2, compras_especiales_after_match_2, compras_especiales_after_match_2_sin_linea_venta = match_2_results_func(match_2, ventas_after_match_1, compras_especiales_after_match_1)

    match_3, match_3_to_errase = match_3_merge_func(ventas_after_match_2, compras_especiales_after_match_2)
    ventas_after_match_3, compras_especiales_after_match_3, compras_especiales_after_match_3_sin_linea_venta = match_3_results_func(match_3_to_errase, ventas_after_match_2, compras_especiales_after_match_2)

    match_4_merge = match_4_merge_func(ventas_after_match_3, compras_especiales_after_match_3)
    match_4, ventas_after_match_4, compras_especiales_after_match_4, compras_especiales_after_match_4_sin_linea_venta = match_4_results_func(match_4_merge, ventas_after_match_3, compras_especiales_after_match_3)

    match_compras_especiales = match_compras_especiales_func(match_1, match_1_repetidos, match_2, match_3, match_4)
    costo_ventas_a = costo_ventas_a_func(match_compras_especiales, ventas, total_compras_odoo)
    costo_ventas_b = costo_ventas_b_func(ventas_after_match_4, compras_no_especiales, ultimo_costo)
    costo_ventas = costo_ventas_func(costo_ventas_a, costo_ventas_b)

    # Costo de venta con las correcciones
    costo_ventas_corregido = costo_ventas_corregido_func(correcciones, compras_line_id_corrected, costo_ventas)

    # Función de checks para detectar erroes antes de generar dataframes finales
    all_checks_costo_ventas = checks_costo_ventas_func(
                                    mes,
                                    compras_doc,
                                    prov_locales,
                                    prov_oficiales,
                                    compras_especiales_after_match_4,
                                    match_compras_especiales,
                                    match_3_to_errase,
                                    match_3,
                                    compras_especiales_after_match_1_sin_linea_venta,
                                    compras_especiales_after_match_2_sin_linea_venta,
                                    compras_especiales_after_match_3_sin_linea_venta,
                                    compras_especiales_after_match_4_sin_linea_venta,
                                    compras_especiales,
                                    costo_ventas_corregido,
                                    ventas
                                )

    if not all_checks_costo_ventas:
        return None, None, None
    
    else:
        # Costo de venta corregido y con las comisiones de los clientes
        costo_ventas_after_comisiones_clientes = costo_ventas_after_comisiones_clientes_func(costo_ventas_corregido, clientes_comisiones)

        # Preparar compras especiales sin usar y último costo para el siguiente mes
        nuevas_compras_especiales_sin_usar = nuevas_compras_especiales_sin_usar_func(compras_especiales_after_match_1_sin_linea_venta, compras_especiales_after_match_2_sin_linea_venta, compras_especiales_after_match_3_sin_linea_venta, compras_especiales_after_match_4_sin_linea_venta, compras_odoo)
        nuevo_ultimo_costo = nuevo_ultimo_costo_func(ultimo_costo, compras_especiales, compras_no_especiales)

    return costo_ventas_after_comisiones_clientes, nuevas_compras_especiales_sin_usar, nuevo_ultimo_costo



import os
import xmlrpc.client
import pandas as pd
from datetime import datetime, timedelta


def api_params_func(test_db: bool = False) -> dict:

    api_url = os.environ.get('ODOO_URL_API')
    api_db = os.environ.get('ODOO_DB_API')
    api_test_db = os.environ.get('ODOO_DB_PRUEBA_API')
    api_username = os.environ.get('ODOO_USERNAME_API')
    api_clave = os.environ.get('ODOO_CLAVE_API')


    api_params = {}
    if test_db:
        api_params['api_db'] = api_test_db
    else:
        api_params['api_db'] = api_db


    common = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/common')
    uid = common.authenticate(api_params['api_db'], api_username, api_clave, {})
    models = xmlrpc.client.ServerProxy(f'{api_url}/xmlrpc/2/object')


    api_params['api_clave'] = api_clave
    api_params['api_uid'] = uid
    api_params['api_models'] = models

    return api_params


def if_list_gt0_idex (item: dict, key: str) -> None | int:
        val = item[key]

        if val:
            if len(val) == 0:
                return None
            else:
                return val[0]
        else:
            return None


def search_fact_func(mes: int, dias: None | list[int] = None, vendedor: None | int  = None) -> list[str]:
    
    if type(mes) != int or mes < 1 or mes > 12:
        raise Exception (f'El mes es incorrecto. El párametro "mes" debe ser un número entero entre 1 y 12. Escribiste: {mes}')
    

    if dias is None:
        param_dia_ini = datetime(2024, mes, 1)
        param_dia_fin = datetime(2024, mes + 1, 1) - timedelta(days= 1)
    
    elif len(dias) != 2:
        raise Exception (f'El párametro "días_del_mes" debe ser una lista de 2 elementos. El día inicial de búsqueda se escribe en el índice 0 de la lista, el día final de búsqueda en el índice 1. La lista tiene: {len(dias)} de elementos')
    
    elif type(dias[0]) != int or type(dias[1]) != int:
        raise Exception (f'El párametro "día_inicial_de_búsqueda" y "día_final_de_búsqueda" sólo pueden ser números enteros.')

    elif dias[0] > dias[1]:
        raise Exception (f'El párametro "día_inicial_de_búsqueda" no debe ser mayor al parámetro "día_final_de_búsqueda". El día inicial de búsqueda se escribe en el índice 0 de la lista, el día final de búsqueda en el índice 1.')
    
    else:
        try:
            param_dia_ini = datetime(2024, mes, dias[0])
        except:
            raise Exception (f'El párametro "día_inicial_de_búsqueda" es incorrecto. La fecha "día: {dias[0]}, mes: {mes}, año 2024" no existe')
        try:
            param_dia_fin = datetime(2024, mes, dias[1])
        except:
            raise Exception (f'El párametro "día_final_de_búsqueda" es incorrecto. La fecha "día: {dias[1]}, mes: {mes}, año 2024" no existe')


    search_fact_all = [
        "&",
            ("state", "=", "posted"),
        "&", "&",
            ("invoice_date", ">=", param_dia_ini.strftime('%Y-%m-%d')),
            ("invoice_date", "<=", param_dia_fin.strftime('%Y-%m-%d')),
            ("journal_id", "in", [10, 90, 30, 97])
        ]

    search_fact_vendedor = [
        "&",
            ("state", "=", "posted"),
        "&", "&",
            ("invoice_date", ">=", param_dia_ini.strftime('%Y-%m-%d')),
            ("invoice_date", "<=", param_dia_fin.strftime('%Y-%m-%d')),
        "&",
            ("journal_id", "in", [10, 90, 30, 97]),
        "|",
            ("invoice_user_id", "=", vendedor),
            ("pos_order_ids.lines.sale_order_origin_id.user_id", "=", vendedor)]


    if vendedor == None:
        return search_fact_all
    
    if type(vendedor) != int:
        raise Exception (f'El párametro "Vendedor" debe ser un número entero. Escribiste: {vendedor}')
    
    return search_fact_vendedor


def api_call_func(api_params: dict, search_fact: list ) -> list[int | dict]:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']


    search_fact = search_fact


    fact_doc_fields = [
                    'name',
                    'invoice_date',
                    'state',
                    'reversed_entry_id',
                    'reversal_move_id',
                    'journal_id',
                    'company_id',
                    'invoice_origin',
                    'pos_order_ids',
                    'line_ids',
                    'partner_id',
                    'move_type',
                    'invoice_user_id'
                    ]

    fact_doc_ids1 = models.execute_kw(api_db, uid, api_clave, 'account.move', 'search', [search_fact])
    fact_doc_ids2 = models.execute_kw(api_db, uid, api_clave, 'account.move', 'search', [[("reversal_move_id", "=", fact_doc_ids1)]])
    fact_doc_ids = list(set(fact_doc_ids1) | set(fact_doc_ids2))
    fact_doc_ids.sort()

    fact_doc_json = models.execute_kw(api_db, uid, api_clave, 'account.move', 'read', [fact_doc_ids], {'fields': fact_doc_fields})

    fact_line_ids = []
    pos_doc_ids1 = []
    invoice_origin_names = []

    for fact in fact_doc_json:
        if fact['pos_order_ids']:
            pos_doc_ids1.append(fact['pos_order_ids'][0])
        
        if fact['invoice_origin']:
            if fact['invoice_origin'][:2] in ['Pd', 'Sh']:
                invoice_origin_names.append(fact['invoice_origin'])

        fact_line_ids += fact['line_ids']


    fact_line_fields = [
            'product_id',
            'quantity',
            'price_unit',
            'discount',
            'account_id',
            'price_subtotal',
        ]


    fact_line_ids.sort()
    fact_line_json = models.execute_kw(api_db, uid, api_clave, 'account.move.line', 'read', [fact_line_ids], {'fields': fact_line_fields})


    pos_line_fields = [
        'name',
        'order_id',
        'sale_order_origin_id',
        'refunded_orderline_id',
    ]

    pos_doc_ids2 = models.execute_kw(api_db, uid, api_clave, 'pos.order', 'search', [[("name", "in", invoice_origin_names)]])
    pos_doc_ids = list(set(pos_doc_ids1) | set(pos_doc_ids2))
    pos_doc_ids.sort()

    pos_line_ids1 = models.execute_kw(api_db, uid, api_clave, 'pos.order.line', 'search', [[("order_id.id", "=", pos_doc_ids)]])
    pos_line_ids2 = models.execute_kw(api_db, uid, api_clave, 'pos.order.line', 'search', [[("refund_orderline_ids", "=", pos_line_ids1)]])
    pos_line_ids = list(set(pos_line_ids1) | set(pos_line_ids2))
    pos_line_ids.sort()
    pos_line_json = models.execute_kw(api_db, uid, api_clave, 'pos.order.line', 'read', [pos_line_ids], {'fields': pos_line_fields})


    sale_doc_ids_each_line = []
    for pos in pos_line_json:
        if pos['sale_order_origin_id']:
            sale_doc_ids_each_line.append(pos['sale_order_origin_id'][0])

    sale_doc_ids = list(set(sale_doc_ids_each_line))
    sale_doc_ids.sort()


    sale_doc_fields = [
        'user_id',
    ]

    sale_doc_json = models.execute_kw(api_db, uid, api_clave, 'sale.order', 'read', [sale_doc_ids], {'fields': sale_doc_fields})

    return fact_doc_ids1, fact_doc_json, fact_line_json, pos_line_json, sale_doc_json


def fact_doc_df_fun(fact_doc_ids: list[int], fact_doc_json: list[dict]) -> pd.DataFrame:
    
    data_fact = []

    for fact in fact_doc_json:
        if fact['id'] in fact_doc_ids:
            for line in fact['line_ids']:
                new = {}
                new['fact_doc_id'] = fact['id']
                new['name'] = fact['name']
                new['invoice_date'] = fact['invoice_date']
                new['state'] = fact['state']
                new['invoice_origin'] = fact['invoice_origin']
                new['module_origin'] = None
                new['pos_doc_id'] = if_list_gt0_idex(fact, 'pos_order_ids')
                new['move_type'] = fact['move_type']
                new['reversal_move_id'] = if_list_gt0_idex(fact, 'reversal_move_id')
                new['reversed_entry_id'] = if_list_gt0_idex(fact, 'reversed_entry_id')
                new['journal_id'] = fact['journal_id'][0]
                new['company_id'] = fact['company_id'][0]
                new['partner_id'] = fact['partner_id'][0]
                new['invoice_user_id'] = fact['invoice_user_id'][0]
                new['fact_line_id'] = line

                if not fact['invoice_origin']:
                    new['module_origin'] = 'Contabilidad'
                elif fact['invoice_origin'][:2] in ['Pd', 'Sh']:
                        new['module_origin'] = 'PdV'
                elif fact['invoice_origin'][0] == 'S':
                        new['module_origin'] = 'Ventas'

                data_fact.append(new)



    fact_doc_df = pd.DataFrame(data_fact)


    fact_doc_df.loc[fact_doc_df['invoice_origin'] == False , ['invoice_origin',]] = pd.NA
    fact_doc_df['invoice_date'] = pd.to_datetime(fact_doc_df['invoice_date'], format='%Y-%m-%d')
    
    for col in fact_doc_df.columns:
         if fact_doc_df[col].dtype in ['int64', 'float64']:
              fact_doc_df[col] = fact_doc_df[col].astype('Int64')

    return fact_doc_df


def fact_line_df_fun(fact_line_json: list[dict]) -> pd.DataFrame:

    data_line_fact = []

    for fact_line in fact_line_json:
        if fact_line['account_id'] and fact_line['account_id'][0] in [85, 197]:
            new = {}
            new['fact_line_id'] = fact_line['id']
            new['product_id'] = fact_line['product_id'][0]
            
            pos = fact_line['product_id'][1].find(']')
            if pos == -1:
                new['product_name'] = fact_line['product_id'][1]
            else:
                new['product_name'] = fact_line['product_id'][1][pos+2 :]
            
            new['quantity'] = fact_line['quantity']
            new['price_unit'] = fact_line['price_unit']
            new['discount'] = fact_line['discount'] / 100
            new['price_subtotal'] = fact_line['price_subtotal']

            data_line_fact.append(new)


    fact_line_df = pd.DataFrame(data_line_fact)


    fact_line_df.loc[fact_line_df['product_id'] == False, ['product_id',]] = pd.NA
    fact_line_df['fact_line_id'] = fact_line_df['fact_line_id'].astype('Int64')
    fact_line_df['product_id'] = fact_line_df['product_id'].astype('Int64')
    
    return fact_line_df


def pos_line_df_fun(pos_line_json: list[dict]) -> pd.DataFrame:

    data_pos_line = []

    for pos in pos_line_json:
        new = {}
        new['pos_line_id'] = pos['id']
        new['pos_doc_id'] = pos['order_id'][0]
        new['sale_doc_id'] = if_list_gt0_idex(pos, 'sale_order_origin_id')
        new['refunded_orderline_id'] = if_list_gt0_idex(pos, 'refunded_orderline_id')
        

        data_pos_line.append(new)

    pos_line_df = pd.DataFrame(data_pos_line)
    
    for col in pos_line_df.columns:
         if pos_line_df[col].dtype in ['int64', 'float64']:
              pos_line_df[col] = pos_line_df[col].astype('Int64')

    
    # Para copiar el 'sale_doc_id' de la línea origial de venta a la línea de devolución que no tiene.
    for id in pos_line_df.loc[(pos_line_df['sale_doc_id'].isna()) & (~pos_line_df['refunded_orderline_id'].isna())]['refunded_orderline_id'].unique():
        pos_line_df.loc[(pos_line_df['refunded_orderline_id'] == id), 'sale_doc_id'] = pos_line_df.loc[(pos_line_df['pos_line_id'] == id), 'sale_doc_id'].iloc[0]

    # Para copiar el 'sale_doc_id' de una línea hermana con el mismo pos_doc_id.
    pos_line_id_not_sale_doc, pos_doc_id_not_sale_doc = pos_line_df.loc[pos_line_df['sale_doc_id'].isna(), ['pos_line_id', 'pos_doc_id']].items()
    for i in range(len(pos_line_id_not_sale_doc[1])):
        pos_line_df.loc[(pos_line_df['pos_line_id'] == pos_line_id_not_sale_doc[1].iloc[i]), 'sale_doc_id'] = pos_line_df.loc[(pos_line_df['pos_doc_id'] == pos_doc_id_not_sale_doc[1].iloc[i]), 'sale_doc_id'].iloc[0]


    return pos_line_df


def sale_doc_df_fun(sale_doc_json: list[dict]) -> pd.DataFrame:

    data_sale_doc = []

    for sale in sale_doc_json:
        new = {}
        new['sale_doc_id'] = sale['id']
        new['salesperson_id'] = sale['user_id'][0]

        data_sale_doc.append(new)


    sale_doc_df = pd.DataFrame(data_sale_doc)
    for col in sale_doc_df.columns:
        sale_doc_df[col] = sale_doc_df[col].astype('Int64')

    return sale_doc_df


def fact_df_func(fact_doc_df: pd.DataFrame, fact_line_df: pd.DataFrame) -> pd.DataFrame:

    fact_df = fact_doc_df.merge(fact_line_df, how='left', on='fact_line_id')
    fact_df.loc[fact_df['move_type'] == 'out_refund', ['quantity', 'price_subtotal']] = fact_df.loc[fact_df['move_type'] == 'out_refund', ['quantity', 'price_subtotal']] * -1
    index_to_drop = fact_df[fact_df['product_id'].isna()].index
    fact_df.drop(index_to_drop, inplace=True)

    return fact_df


def pos_sale_df_func(pos_line_df: pd.DataFrame, sale_doc_df: pd.DataFrame) -> list[pd.DataFrame]:
    pos_sale_df = pos_line_df.merge(sale_doc_df, how='left', on='sale_doc_id')

    saleperson_FvsM = pos_sale_df.loc[:, ['pos_doc_id', 'salesperson_id']].groupby('pos_doc_id').agg({'salesperson_id':['first', 'mean']})
    saleperson_FvsM['is_different'] = saleperson_FvsM[('salesperson_id', 'first')] != saleperson_FvsM[('salesperson_id', 'mean')]

    pos_doc_equal_ids_df = saleperson_FvsM.loc[saleperson_FvsM['is_different'] == False, [('salesperson_id', 'first')]].reset_index()
    pos_doc_equal_ids_df.columns = ['pos_doc_id', 'salesperson_id']

    pos_doc_different_ids = ((saleperson_FvsM.loc[saleperson_FvsM['is_different'] == True])).reset_index()['pos_doc_id']
    pos_doc_different_ids_df = pos_sale_df.loc[pos_sale_df['pos_doc_id'].isin(pos_doc_different_ids), ['pos_line_id', 'pos_doc_id', 'salesperson_id']]

    return pos_doc_equal_ids_df, pos_doc_different_ids_df


def complete_df_func(api_params: dict, fact_df: pd.DataFrame, pos_doc_equal_ids_df: pd.Series, pos_doc_different_ids_df: pd.Series) -> pd.DataFrame:
    
    api_db = api_params['api_db']
    api_clave = api_params['api_clave']
    uid = api_params['api_uid']
    models = api_params['api_models']


    # Se complementa el campo 'salesperson_id' de todos los fact_doc_id que tienen en su contraparte pos_doc_id un mismo vendedor
    complete_df = fact_df.merge(pos_doc_equal_ids_df, how='left', on='pos_doc_id')
    
    
    # Se complementa el campo 'salesperson_id' de todos los fact_doc_id que provienen de los módulos de origen 'Contabilidad' y 'Ventas'
    complete_df.loc[complete_df['module_origin'] != 'PdV', 'salesperson_id'] = complete_df.loc[complete_df['module_origin'] != 'PdV', 'invoice_user_id']
    

    # Se complementa el campo 'salesperson_id' de todos los fact_doc_id que tienen en su contraparte pos_doc_id un vendedor diferente en alguna de sus líneas
    if not pos_doc_different_ids_df.empty:
        for id in pos_doc_different_ids_df['pos_doc_id'].unique():
            if len(complete_df.loc[complete_df['pos_doc_id'] == id]) == len(pos_doc_different_ids_df.loc[pos_doc_different_ids_df['pos_doc_id'] == id]):
                complete_df.loc[complete_df['pos_doc_id'] == id, 'salesperson_id'] = list(pos_doc_different_ids_df.loc[pos_doc_different_ids_df['pos_doc_id'] == id, 'salesperson_id'])
    
    
    # Se complementa al campo 'salesperson_id' de las líneas tipo 'out_refund' que no tienen un pos_doc_id. Se trata de buscar la línea de la factura original que coincide con el producto para copiar el 'salesperson_id' a la línea de devolución.
    fact_line_id_not_saleline, reversed_entry_id_not_saleline, product_id_not_saleline = complete_df.loc[(complete_df['salesperson_id'].isna()) & (complete_df['move_type'] == 'out_refund'), ['fact_line_id', 'reversed_entry_id', 'product_id']].items()

    for i in range(len(fact_line_id_not_saleline[1])):
        if not complete_df.loc[complete_df['fact_doc_id'] == reversed_entry_id_not_saleline[1].iloc[i]].empty:
            complete_df.loc[complete_df['fact_line_id'] == fact_line_id_not_saleline[1].iloc[i], 'salesperson_id'] = complete_df.loc[(complete_df['fact_doc_id'] == reversed_entry_id_not_saleline[1].iloc[i]) & (complete_df['product_id'] == product_id_not_saleline[1].iloc[i]), 'salesperson_id'].iloc[0]

    
    # Se complementa al campo 'salesperson_id' de las líneas restantes. No tienen un pos_doc_id y el 'invoice_origin' no está en los json. Ha que llamar a la API_Odoo de nuevo.
    pos_doc_names_not_sale_doc, product_id_not_sale_doc = complete_df.loc[complete_df['salesperson_id'].isna(), ['invoice_origin', 'product_id']].items()

    pos_line_not_found_json = models.execute_kw(api_db, uid, api_clave, 'pos.order.line', 'search_read', [["&", ("order_id.name", "in", list(pos_doc_names_not_sale_doc[1])), ("product_id", "in", list(product_id_not_sale_doc[1].astype(int)))]], {'fields': ['order_id', 'product_id', 'sale_order_line_id']})

    pos_line_not_found_ids = []
    for pos in pos_line_not_found_json:
        if pos['sale_order_line_id']:
            pos_line_not_found_ids.append( pos['sale_order_line_id'][0])

    sale_pos_line_not_found_json = models.execute_kw(api_db, uid, api_clave, 'sale.order.line', 'read', [pos_line_not_found_ids], {'fields': ['salesman_id']})

    for i in range(len(pos_doc_names_not_sale_doc[1])):
        for pos in pos_line_not_found_json:
            if pos_doc_names_not_sale_doc[1].iloc[i] == pos['order_id'][1] and product_id_not_sale_doc[1].iloc[i] == pos['product_id'][0]:
                for sale in sale_pos_line_not_found_json:
                    if pos['sale_order_line_id']:
                        if pos['sale_order_line_id'][0] == sale['id']:
                            complete_df.loc[(complete_df['salesperson_id'].isna()) & (complete_df['invoice_origin'] == pos_doc_names_not_sale_doc[1].iloc[i]) & (complete_df['product_id'] == product_id_not_sale_doc[1].iloc[i]), 'salesperson_id'] = sale['salesman_id'][0]
            
    return complete_df


def ventas_mes_func(mes: int, dias: None | list[int] = None, vendedor: None | int  = None, test_db: bool = False) -> pd.DataFrame:
    api_params = api_params_func(test_db)
    search_fact = search_fact_func(mes, dias, vendedor)
    
    fact_doc_ids, fact_doc_json, fact_line_json, pos_line_json, sale_doc_json = api_call_func(api_params, search_fact)
    fact_doc_df = fact_doc_df_fun(fact_doc_ids, fact_doc_json)
    fact_line_df = fact_line_df_fun(fact_line_json)
    pos_line_df = pos_line_df_fun(pos_line_json)
    sale_doc_df = sale_doc_df_fun(sale_doc_json)

    fact_df = fact_df_func(fact_doc_df, fact_line_df)
    pos_doc_equal_ids_df, pos_doc_different_ids_df = pos_sale_df_func(pos_line_df, sale_doc_df)

    complete_df = complete_df_func(api_params, fact_df, pos_doc_equal_ids_df, pos_doc_different_ids_df)

    return complete_df



"""
## IACele
Librería para crear BI en la administración de la empresa LA CASA DEL CARPINTERO.
"""

from .algoritmo_ventas_mes import ventas_mes_func
from .algoritmo_cobranza_mes import cobranza_mes_func
from .algoritmo_costo_ventas_mes import costo_ventas_mes_func